﻿using System;
using System.Collections.Generic;
using System.Linq;

namespace HPALMRestAPIWrapper
{
    /// <summary>
    /// Defect collection class
    /// </summary>
    public class ALMDefects : ALMEntityCollection<ALMDefect>
    {
        #region Constructor
        /// <summary>
        /// constructor
        /// </summary>
        /// <param name="connection">connection</param>
        public ALMDefects(ALMConnection connection)
            : base(connection)
        {}
        #endregion

        #region public methods
        /// <summary>
        /// Create Defect
        /// </summary>
        /// <param name="defectFields">Defect fields</param>
        /// <returns>Defect</returns>
        public override ALMDefect Create(Dictionary<string, string> defectFields)
        {
            Common.SetDictionaryValue(ref defectFields, "detected-by", Connection.UserName);
            Common.SetDictionaryValue(ref defectFields, "creation-time", DateTime.UtcNow.ToString("yyyy-MM-dd"));
            Common.SetDictionaryValue(ref defectFields, "status", "New");

            return base.Create(defectFields);
        }

        /// <summary>
        /// Set defect status to Open 
        /// </summary>
        /// <param name="id">Defect Id</param>
        /// <returns>Defect</returns>
        public ALMDefect Open(int id)
        {
            return ChangeStatus(id, "Open");
        }

        /// <summary>
        /// Set defect status to Fixed 
        /// </summary>
        /// <param name="id">Defect Id</param>
        /// <returns>Defect</returns>
        public ALMDefect Fix(int id)
        {
            return ChangeStatus(id, "Fixed");
        }

        /// <summary>
        /// Set defect status to Rejected 
        /// </summary>
        /// <param name="id">Defect Id</param>
        /// <returns>Defect</returns>
        public ALMDefect Reject(int id)
        {
            return ChangeStatus(id, "Rejected");
        }

        /// <summary>
        /// Set defect status to Closed 
        /// </summary>
        /// <param name="id">Defect Id</param>
        /// <returns>Defect</returns>
        public ALMDefect Close(int id)
        {
            return ChangeStatus(id, "Closed");
        }

        /// <summary>
        /// Set defect status to Reopen 
        /// </summary>
        /// <param name="id">Defect Id</param>
        /// <returns>Defect</returns>
        public ALMDefect Reopen(int id)
        {
            return ChangeStatus(id, "Reopen");
        }

        /// <summary>
        /// Search defects
        /// </summary>
        /// <param name="filterCondition">search condition</param>
        /// <returns></returns>
        public List<ALMDefect> Search(Dictionary<string, string> filterCondition)
        {
            return base.GetCollection(filterCondition);
        }
        #endregion

        #region private methods

        private ALMDefect ChangeStatus(int id, string status)
        {
            var fields = new Dictionary<string, string> {{"status", status}};

            return base.Update(id, fields);
        }
        #endregion
    }

    /// <summary>
    /// Defect class
    /// </summary>
    public class ALMDefect: ALMEntity
    {
        #region private members
        #endregion

        #region public Properties
        #endregion

        #region constructor

        #endregion

        #region public methods
        /// <summary>
        /// Update defect
        /// </summary>
        /// <returns></returns>
        public ALMDefect Update()
        {
            if (Id <= 0) throw new Exception("Bug Id cannot be empty.");

            return new ALMDefects(Connection).Update(Id.Value, GetFieldsDictionary());
        }

        /// <summary>
        /// Open defect
        /// </summary>
        /// <returns></returns>
        public ALMDefect Open()
        {
            SetUpdateField("status", "Open");
            return Update();
        }

        /// <summary>
        /// Fix defect
        /// </summary>
        /// <returns></returns>
        public ALMDefect Fix()
        {
            SetUpdateField("status", "Fixed");
            return Update();
        }

        /// <summary>
        /// Reject defect
        /// </summary>
        /// <returns></returns>
        public ALMDefect Reject()
        {
            SetUpdateField("status", "Rejected");
            return Update();
        }

        /// <summary>
        /// Close defect
        /// </summary>
        /// <returns></returns>
        public ALMDefect Close()
        {
            SetUpdateField("status", "Closed");
            return Update();
        }

        /// <summary>
        /// Reopen defect
        /// </summary>
        /// <returns></returns>
        public ALMDefect Reopen()
        {
            SetUpdateField("status", "Reopen");
            return Update();
        }

        #region Links in defect
        /// <summary>
        /// link to defects
        /// </summary>
        /// <param name="defectIds">defect ids</param>
        public void LinkToDefects(int[] defectIds)
        {
            if (defectIds == null)
                throw new Exception("Defect Ids are empty.");

            var defectlinks = new ALMDefectLinks(Connection);

            foreach (var defectId in defectIds)
            {
                defectlinks.Create(defectId, Id.Value, "defect");
            }
        }

        /// <summary>
        /// link to requirements
        /// </summary>
        /// <param name="reqIds">requirement Ids</param>
        public void LinkToRequirements(int[] reqIds)
        {
            if (reqIds == null)
                throw new Exception("Requirements Ids are empty.");

            var defectlinks = new ALMDefectLinks(Connection);

            foreach (var reqId in reqIds)
            {
                defectlinks.Create(Id.Value, reqId, "requirement");
            }
        }

        /// <summary>
        /// Get linked defects
        /// </summary>
        /// <returns>defect list</returns>
        public List<ALMDefect> GetLinkedDefects()
        {
            var linkeddefects = new List<ALMDefect>();

            var defectlinks = new ALMDefectLinks(Connection);
            var filters = new Dictionary<string, string>
                              {
                                  {"second-endpoint-id", Id.Value.ToString()},
                                  {"second-endpoint-type", "defect"}
                              };
            var dllist = defectlinks.GetCollection(filters);

            if (dllist != null)
            {
                var defects = new ALMDefects(Connection);
                linkeddefects.AddRange(dllist.Select(dl => defects.Get(Convert.ToInt32(dl.GetField("second-endpoint-id").Value))).Where(d => d != null));
            }

            return linkeddefects;
        }

        /// <summary>
        /// Get linked requirements
        /// </summary>
        /// <returns>requirement list</returns>
        public List<ALMRequirement> GetLinkedRequirements()
        {
            var linkedrequirements = new List<ALMRequirement>();

            var defectlinks = new ALMDefectLinks(Connection);
            var filters = new Dictionary<string, string>
                              {
                                  {"first-endpoint-id", Id.Value.ToString()},
                                  {"second-endpoint-type", "requirement"}
                              };
            var dllist = defectlinks.GetCollection(filters);

            if (dllist != null)
            {
                var reqs = new ALMRequirements(Connection);
                linkedrequirements.AddRange(dllist.Select(dl => reqs.Get(Convert.ToInt32(dl.GetField("second-endpoint-id").Value))).Where(d => d != null));
            }

            return linkedrequirements;
        }

        /// <summary>
        /// Delete defect links
        /// </summary>
        /// <param name="defectIds">defect ids</param>
        public void DeleteDefectLinks(int[] defectIds)
        {
            if (defectIds == null)
                throw new Exception("Defect Ids are empty.");

            var defectlinks = new ALMDefectLinks(Connection);

            foreach (var dl in
                defectIds.Select(defectId => defectlinks.Get(defectId, Id.Value, "defect")).Where(dl => dl != null))
            {
                dl.Delete();
            }
        }

        /// <summary>
        /// delete requirement links
        /// </summary>
        /// <param name="reqIds">requirement ids</param>
        public void DeleteRequirementLinks(int[] reqIds)
        {
            if (reqIds == null)
                throw new Exception("Requirement Ids are empty.");

            var defectlinks = new ALMDefectLinks(Connection);

            foreach (var dl in
                reqIds.Select(reqId => defectlinks.Get(Id.Value, reqId, "requirement")).Where(dl => dl != null))
            {
                dl.Delete();
            }
        }
        #endregion

        #endregion
    }
}
