﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Xml;

namespace HPALMRestAPIWrapper
{
    /// <summary>
    /// Abstract class for ALMEntity Collection, be inherited by each Entity class, for example: Defect, Test, Requirement
    /// </summary>
    /// <typeparam name="T"></typeparam>
    public abstract class ALMEntityCollection<T> where T : ALMEntity, new()
    {
        #region Priavte Members
        #endregion

        #region Public Properties
        /// <summary>
        /// Connection object of ALM connection
        /// </summary>
        public ALMConnection Connection { get; set; }
        /// <summary>
        /// Entity fields object
        /// </summary>
        public List<ALMCustomizedField> CustomizedFields { get; set; }
        /// <summary>
        /// Types object
        /// </summary>
        public List<ALMCustomizedType> CustomizedTypes { get; set; }
        /// <summary>
        /// Entity Rest API URL
        /// </summary>
        public string RestUrl { get; set; }
        /// <summary>
        /// Entity type: defect, requirement, test, run, defect-link, test-set, test-folder, testset-folder
        /// </summary>
        public string EntityType { get; set; }
        #endregion

        #region constructor
        /// <summary>
        /// Entity Collection
        /// </summary>
        /// <param name="connection"></param>
        public ALMEntityCollection(ALMConnection connection)
        {
            Connection = connection;
            EntityType = Common.GetEntityType(typeof(T).Name).Trim().Replace("\"", "").Substring(0, Common.GetEntityType(typeof(T).Name).Trim().Length - 1); // Get entity type, for example: defect/test/run
            RestUrl = string.Format("http://{0}:{1}/qcbin/rest/domains/{2}/projects/{3}/{4}s", connection.ServerName, connection.ServerPort, connection.Domain, connection.Project, EntityType); // Get Rest URL for different entity type.
            CustomizedFields = ALMCustomizedField.GetFields(connection, EntityType); // Get all fields definitions from different entity type, each project or version with different fields definitions.
            CustomizedTypes = ALMCustomizedType.GetTypes(connection, EntityType); // Get all types definition from different entity type, not each entity with sub-types.
        }
        #endregion        

        #region public methods
        /// <summary>
        /// Get entity object by its ID
        /// </summary>
        /// <param name="entityId">Entity ID</param>
        /// <param name="fields">Return Fields for the entity</param>
        /// <returns>Entity</returns>
        public virtual T Get(int entityId, params string[] fields)
        {
            var urlSb = new StringBuilder();
            try
            {
                #region Build Query string and Field string
                var fieldsStr = Common.BuildFields(fields);
                urlSb.Append(string.Format("{0}/{1}", RestUrl, entityId));
                if (!string.IsNullOrEmpty(fieldsStr))
                    urlSb.Append(string.Format("{0}?{1}", urlSb, fieldsStr));
                #endregion

                #region Request to get entity
                Connection.Http.Method = "GET";
                Connection.Http.ContentType = "application/xml";
                var status = Connection.Http.Request(urlSb.ToString());
                if (status >= 200 && status <= 204)
                {
                    if (!string.IsNullOrEmpty(Connection.Http.Response))
                    {
                        var doc = new XmlDocument();
                        doc.LoadXml(Connection.Http.Response);

                        return LoadEntityFromXml(doc.SelectSingleNode("Entity/Fields").OuterXml);
                    }
                }
                #endregion

                return null;
            }
            catch (Exception ex)
            {
                throw new Exception(ex.Message);
            }
        }

        /// <summary>
        /// Get entity collection by the parent Id
        /// </summary>
        /// <param name="parentId">Parent Id for entity collection</param>
        /// <param name="fields">Return Fields for the entity</param>
        /// <returns>List of Entity</returns>
        public virtual List<T> GetCollection(int? parentId, params string[] fields)
        {
            if (parentId == null)  throw new Exception("Parent Id for Entity cannot be empty.");
            if (CustomizedFields.Where(o=>o.Name.Equals("parent-id", StringComparison.InvariantCultureIgnoreCase)).Count() != 1) 
                throw new Exception("The entity type without the field 'parent-id'."); // Check if there is a field which named 'parent-id' in the entity fields definitions, if not, try second GetCollection methods.

            var filters = new Dictionary<string, string> {{"parent-id", parentId.ToString()}};

            return GetCollection(filters, fields);
        }

        /// <summary>
        /// Get entity collection by the filter conditions
        /// </summary>
        /// <param name="filters">Filter conditions for the entity collection</param>
        /// <param name="fields">Return Fields for the entity</param>
        /// <returns>List of Entity</returns>
        public virtual List<T> GetCollection(Dictionary<string, string> filters, params string[] fields)
        {
            if (filters.Count == 0) throw new Exception("Filters for Entity cannot be empty.");

            var collection = new List<T>();
            var urlSb = new StringBuilder();

            try
            {
                CheckFields(filters);
                #region Build Query string and Field string
                var filtersStr = Common.BuildFilters(NormalizeFields(filters));
                var fieldsStr = Common.BuildFields(fields);

                urlSb.Append(string.Format("{0}/?", RestUrl));
                if (!string.IsNullOrEmpty(filtersStr))
                    urlSb.Append(string.Format("{0}&", filtersStr));

                if (!string.IsNullOrEmpty(fieldsStr))
                    urlSb.Append(string.Format("{0}", fieldsStr));
                #endregion

                #region Request to Get Collection
                Connection.Http.Method = "GET";
                Connection.Http.ContentType = "application/xml";
                var status = Connection.Http.Request(urlSb.ToString());

                if (status >= 200 && status <= 204)
                {
                    var response = Connection.Http.Response;
                    if (!string.IsNullOrEmpty(response))
                    {
                        #region Load entity info to object list
                        var doc = new XmlDocument();
                        doc.LoadXml(response);
                        var nodelist = doc.SelectNodes("Entities/Entity");
                        collection.AddRange(from XmlNode node in nodelist
                                            select LoadEntityFromXml(node.SelectSingleNode("Fields").OuterXml)
                                            into t where t.CustomizedFields.Count > 0 select t);

                        #endregion
                    }
                #endregion
                }
            }
            catch(Exception ex)
            {
                throw new Exception(ex.Message);
            }

            return collection;
        }

        /// <summary>
        /// Check entity is exists.
        /// </summary>
        /// <param name="entityId">Entity Id</param>
        /// <returns>true of false</returns>
        public virtual bool IsExists(int entityId)
        {
            var result = false;

            try
            {
                var t = Get(entityId);
                if (t != null)
                    result = true;
            }
            catch
            {
                result = false;
            }

            return result;
        }

        /// <summary>
        /// Create entity
        /// </summary>
        /// <param name="entityFields">Entity Fields which will be used to create.</param>
        /// <returns>New Entity</returns>
        public virtual T Create(Dictionary<string, string> entityFields)
        {
            if (entityFields.Count == 0)   throw new Exception("Entity fields cannot be empty.");

            CheckRequiredFieldsAreReady(entityFields);

            var urlSb = new StringBuilder();
            urlSb.Append(string.Format("{0}", RestUrl));

            try
            {
                var xmlContent = Common.BuildEntityXML(NormalizeFields(entityFields), EntityType);
                
                #region Request to Create Entity
                Connection.Http.Method = "POST";
                Connection.Http.ContentType = "application/xml";
                Connection.Http.Accept = "application/xml";
                Connection.Http.PostData = xmlContent;
                var status = Connection.Http.Request(urlSb.ToString());
                if (status >= 200 && status <= 204)
                {
                    var response = Connection.Http.Response;
                    if (!string.IsNullOrEmpty(response))
                    {
                        var doc = new XmlDocument();
                        doc.LoadXml(response);

                        return LoadEntityFromXml(doc.SelectSingleNode("Entity/Fields").OuterXml);
                    }
                }
                #endregion

                return null;
            }
            catch (Exception ex)
            {
                throw new Exception(ex.Message);
            }
        }

        /// <summary>
        /// Update Entity
        /// </summary>
        /// <param name="entityId">Entity Id which need to be updated.</param>
        /// <param name="entityFields">Entity fields which need to be updated.</param>
        /// <returns>Updated Entity</returns>
        public virtual T Update(int entityId, Dictionary<string, string> entityFields)
        {
            if (entityFields.Count == 0) throw new Exception("Entity fields cannot be empty.");

            var urlSb = new StringBuilder();
            urlSb.Append(string.Format(@"{0}/{1}", RestUrl, entityId));

            try
            {
                var xmlContent = Common.BuildEntityXML(NormalizeFields(entityFields), EntityType);

                #region Request to Update entity
                Connection.Http.Method = "PUT";
                Connection.Http.ContentType = "application/xml";
                Connection.Http.Accept = "application/xml";
                Connection.Http.PostData = xmlContent;
                var status = Connection.Http.Request(urlSb.ToString());

                if (status >= 200 && status <= 204)
                {
                    var response = Connection.Http.Response;
                    if (!string.IsNullOrEmpty(response))
                    {
                        var doc = new XmlDocument();
                        doc.LoadXml(response);

                        return LoadEntityFromXml(doc.SelectSingleNode("Entity/Fields").OuterXml);
                    }
                }
                #endregion

                return null;
            }
            catch (Exception ex)
            {
                throw new Exception(ex.Message);
            }
        }

        /// <summary>
        /// Delete Entity
        /// </summary>
        /// <param name="entityId">Entity Id which need to be deleted</param>
        public virtual void Delete(int entityId)
        {
            try
            {
                var query = string.Format(@"{0}/{1}", RestUrl, entityId);

                #region Request to Delete Entity
                Connection.Http.Method = "DELETE";
                Connection.Http.Request(query);
                #endregion
            }
            catch (Exception ex)
            {
                throw new Exception(ex.Message);
            }
        }

        /// <summary>
        /// Find folder by path
        /// </summary>
        /// <param name="folderPath">Folder path, must be start with '\', example: \Requirement\Release1\Create</param>
        /// <returns>Folder Entity</returns>
        public virtual T FindFolderByPath(string folderPath)
        {
            if (string.IsNullOrEmpty(folderPath)) throw new Exception("The folder path cannot be empty.");
            if (!folderPath.StartsWith(@"\"))   throw new Exception(@"The folder path must be start with \ .");

            T tFolder = null;

            var parentId = -1;
            foreach (var foldername in folderPath.Remove(0, 1).Split(new[] { '\\' }))
            {
                tFolder = GetFolder(foldername, parentId);
                if (tFolder == null)
                    throw new Exception(string.Format("Cannot find the folder: {0} ", foldername));

                parentId = tFolder.Id.Value;
            }

            return tFolder;
        }        
        #endregion

        #region private methods
        /// <summary>
        /// Normalize the fields, some field name which be inputted are Field Label, need to be convert to field name.
        /// </summary>
        /// <param name="dic"></param>
        /// <returns></returns>
        private Dictionary<string, string> NormalizeFields(Dictionary<string, string> dic)
        {
            var normalizedFields = new Dictionary<string, string>();
            foreach (var keyvalue in dic)
            {
                if (CustomizedFields.Where(o => o.Name.Equals(keyvalue.Key)).Count() == 1)
                    normalizedFields.Add(keyvalue.Key, keyvalue.Value);
                else if (CustomizedFields.Where(o => o.Label.Equals(keyvalue.Key, StringComparison.InvariantCultureIgnoreCase)).Count() == 1)
                    normalizedFields.Add(CustomizedFields.Where(o => o.Label.Equals(keyvalue.Key, StringComparison.InvariantCultureIgnoreCase)).First().Name, keyvalue.Value);
            }

            return normalizedFields;
        }

        private void CheckRequiredFieldsAreReady(IDictionary<string, string> dic)
        {
            foreach (var almCf in
                CustomizedFields.Where(o => o.IsRequired).Where(almCf => !dic.ContainsKey(almCf.Name) && !dic.ContainsKey(almCf.Label)))
            {
                throw new Exception(string.Format("Field '{0}'[{1}] for '{2}' is required, please fill in.", almCf.Label, almCf.Name, typeof(T).Name));
            }
        }

        private void CheckFields(Dictionary<string, string> dic)
        {
            foreach (var f in
                dic.Where(f => CustomizedFields.Where(o => o.Name.Equals(f.Key, StringComparison.InvariantCultureIgnoreCase)).Count() != 1).Where(f => CustomizedFields.Where(o => o.Label.Equals(f.Key, StringComparison.InvariantCultureIgnoreCase)).Count() != 1))
            {
                throw new Exception(string.Format("The entity type without the field '{0}'.", f.Key));
            }
        }

        private T LoadEntityFromXml(string response)
        {
            var t = new T
                        {
                EntityXml = response,
                RestUrl = RestUrl,
                Connection = Connection,
                CustomizedFields = ALMCustomizedField.GetFields(Connection, EntityType),
            };            
            t.Load();

            return t; 
        }

        private T GetFolder(string foldername, int parentId)
        {
            if (!EntityType.Equals("test-folder") && !EntityType.Equals("testset-folder"))
                throw new Exception("Only Test Folder and TestSet Folder ared supported.");

            var queries = new Dictionary<string, string> {{"name", foldername}};
            if (parentId >= 0)
                queries.Add("parent-id", parentId.ToString());

            CheckFields(queries);

            var folderlist = GetCollection(queries);

            if (folderlist.Count == 0)  throw new Exception(string.Format("Cannot find the folder: {0} ", foldername));
            if (folderlist.Count > 1)  throw new Exception(string.Format("Found multiple folders: {0} ", foldername));

            return folderlist[0];
        }
        #endregion
    }

    /// <summary>
    /// Entity class, be inherited by specified entity, for example: Defect/Requirement/Test
    /// </summary>
    public class ALMEntity
    {
        #region private members
        #endregion

        #region Public properties
        /// <summary>
        /// Store the Rest URL for the entity
        /// </summary>
        public string RestUrl { get; set; }

        /// <summary>
        /// Store the Connection for the entity
        /// </summary>
        public ALMConnection Connection { get; set; }

        /// <summary>
        /// Store the response XML content for the entity.
        /// </summary>
        public string EntityXml { get; set; }

        /// <summary>
        /// Entity Id, when it is null, means for searching operation or create operation.
        /// </summary>
        public int? Id { get; set; }

        /// <summary>
        /// Store all fields and values for the entity
        /// </summary>
        public List<ALMCustomizedField> CustomizedFields { get; set; }
        #endregion

        #region constructor

        #endregion

        #region public method
        #region Attachment Related Methods
        /// <summary>
        /// Upload attachment
        /// </summary>
        /// <param name="file">upload attachment file path</param>
        public void UploadAttachment(string file)
        {
            var query = string.Format(@"{0}/{1}/attachments", RestUrl, Id);

            if (!File.Exists(file))
                throw new Exception(string.Format("File {0} does not existed.", file));

            try
            {
                Connection.Http.PostFilePath = file;
                Connection.Http.Method = "POST";
                Connection.Http.ContentType = "application/octet-stream";
                if (Connection.Http.RequestHeaders.ContainsKey("Slug"))
                    Connection.Http.RequestHeaders["Slug"] = Path.GetFileName(file);
                else
                    Connection.Http.RequestHeaders.Add("Slug", Path.GetFileName(file));
                Connection.Http.Request(query);
            }
            catch (Exception ex)
            {
                throw new Exception(ex.Message);
            }
        }

        /// <summary>
        /// Download attachment from ALM
        /// </summary>
        /// <param name="attachmentName">attachment file name in ALM</param>
        /// <param name="file">local file path</param>
        public void DownloadAttachment(string attachmentName, string file)
        {
            var query = string.Format(@"{0}/{1}/attachments/{2}", RestUrl, Id, attachmentName);

            try
            {
                Connection.Http.DownloadFile(query, file);
            }
            catch (Exception ex)
            {
                throw new Exception(string.Format("Download Attachment Fail! {0}", ex.InnerException.Message));
            }
        }

        /// <summary>
        /// Delete Attachment in ALM
        /// </summary>
        /// <param name="attachmentName">attachment file name</param>
        public void DeleteAttachment(string attachmentName)
        {
            var query = string.Format(@"{0}/{1}/attachments/{2}", RestUrl, Id, attachmentName);

            try
            {
                Connection.Http.Method = "DELETE";
                Connection.Http.ContentType = "application/octet-stream";
                Connection.Http.Request(query);
            }
            catch (Exception ex)
            {
                throw new Exception(ex.Message);
            }
        }

        /// <summary>
        /// Check attachment is exists in ALM
        /// </summary>
        /// <param name="attachmentName">Attachment file name</param>
        /// <returns></returns>
        public bool IsAttachmentExists(string attachmentName)
        {
            var query = string.Format(@"{0}/{1}/attachments/{2}", RestUrl, Id, attachmentName);
            var result = false;

            try
            {
                Connection.Http.Method = "GET";
                Connection.Http.ContentType = "application/octet-stream";
                var status = Connection.Http.Request(query);

                if (status >= 200 && status <= 204)
                    result = true;
            }
            catch (Exception ex)
            {
                if (ex.Message.IndexOf("No entity of type attachment with id") >= 0)
                    result = false;
                else
                    throw new Exception(ex.Message);
            }

            return result;
        }
        #endregion

        #region Common Methods
        /// <summary>
        /// Delete Entity
        /// </summary>
        public void Delete()
        {
            Delete(false);
        }

        /// <summary>
        /// Delete Entity with children
        /// </summary>
        /// <param name="isDeleteChildren">true = delete all chilren</param>
        public void Delete(bool isDeleteChildren)
        {
            if (Id == null)
                throw new Exception("Entity Id cannot be empty.");

            try
            {
                var query = string.Format(@"{0}/{1}", RestUrl, Id);

                if (isDeleteChildren)
                    query = string.Format("{0}?{1}", query, "force-delete-children=y");

                Connection.Http.Method = "DELETE";
                Connection.Http.Request(query);           
            }
            catch (Exception ex)
            {
                throw new Exception(ex.Message);
            }
        }

        /// <summary>
        /// Set fields which need to be updated
        /// </summary>
        /// <param name="fieldname">field name</param>
        /// <param name="value">field value</param>
        public void SetUpdateField(string fieldname, string value)
        {
            if (CustomizedFields.Where(o => o.Name.Equals(fieldname)).Count() == 1)
            {
                CustomizedFields.Where(o => o.Name.Equals(fieldname)).First().Value = value;
                CustomizedFields.Where(o => o.Name.Equals(fieldname)).First().IsToUpdate = true;
            }                
            else if (CustomizedFields.Where(o => o.Label.Equals(fieldname, StringComparison.InvariantCultureIgnoreCase)).Count() == 1)
            {
                CustomizedFields.Where(o => o.Label.Equals(fieldname, StringComparison.InvariantCultureIgnoreCase)).First().Value = value;
                CustomizedFields.Where(o => o.Label.Equals(fieldname, StringComparison.InvariantCultureIgnoreCase)).First().IsToUpdate = true;
            }                
            else
                throw new Exception(string.Format("Cannot find the field '{0}'.", fieldname));
        }

        /// <summary>
        /// Get field by field name
        /// </summary>
        /// <param name="fieldname">field name</param>
        /// <returns></returns>
        public ALMCustomizedField GetField(string fieldname)
        {
            if (CustomizedFields.Where(o => o.Name.Equals(fieldname, StringComparison.InvariantCultureIgnoreCase)).Count() == 1)
                return CustomizedFields.Where(o => o.Name.Equals(fieldname, StringComparison.InvariantCultureIgnoreCase)).First();
            if (CustomizedFields.Where(o => o.Label.Equals(fieldname, StringComparison.InvariantCultureIgnoreCase)).Count() == 1)
                return CustomizedFields.Where(o => o.Label.Equals(fieldname, StringComparison.InvariantCultureIgnoreCase)).First();
            throw new Exception(string.Format("Cannot find the field '{0}'.", fieldname));
        }

        #endregion        

        /// <summary>
        /// Load entity information
        /// </summary>
        public void Load()
        {
            if (string.IsNullOrEmpty(EntityXml)) throw new Exception("Entity Xml is empty.");
            try
            {
                var doc = new XmlDocument();
                doc.LoadXml(EntityXml);

                Id = Convert.ToInt32(Common.GetFieldValue(doc, "id"));
                var fields = doc.SelectNodes("Fields/Field");

                foreach (XmlNode field in fields)
                {
                    var fieldname = field.Attributes["Name"].InnerText;
                    string fieldvalue = null;
                    if (field.HasChildNodes) fieldvalue = field.SelectSingleNode("Value").InnerText;

                    CustomizedFields.Where(o => o.Name.Equals(fieldname)).First().Value = fieldvalue;
                }
            }
            catch (Exception ex)
            {
                throw new Exception(ex.Message);
            }
        }

        /// <summary>
        /// Build dictionary from fields in entity
        /// </summary>
        /// <returns></returns>
        public Dictionary<string, string> GetFieldsDictionary()
        {
            var fields = new Dictionary<string, string>();
            foreach (var almCf in CustomizedFields.Where(o => o.IsToUpdate))
            {
                fields.Add(almCf.Name, almCf.Value);
                almCf.IsToUpdate = false;
            }

            return fields;
        }
        #endregion

        #region Private methods
        #endregion
    }
}
