﻿using System;
using System.Collections.Generic;
using System.Linq;

namespace HPALMRestAPIWrapper
{
    /// <summary>
    /// requirment collection class
    /// </summary>
    public class ALMRequirements : ALMEntityCollection<ALMRequirement>
    {
        /// <summary>
        /// constructor
        /// </summary>
        /// <param name="connection">connection</param>
        public ALMRequirements(ALMConnection connection)
            : base(connection)
        {            
        }

        #region public methods
        /// <summary>
        /// Create Requirement
        /// </summary>
        /// <param name="requirementType">Requirement type</param>
        /// <param name="requirementFields">Requirement fields</param>
        /// <returns>Requirement</returns>
        public ALMRequirement Create(string requirementType, Dictionary<string, string> requirementFields)
        {
            if (string.IsNullOrEmpty(requirementType))
                throw new Exception("Requirement Type cannot be empty.");

            if (CustomizedTypes.Where(o => o.Name.ToLower().Equals(requirementType.ToLower())).Count() == 0)
                throw new Exception(string.Format("Cannot find the requirement type: {0}, please double check", requirementType));

            string typeid = (CustomizedTypes.Where(o => o.Name.ToLower().Equals(requirementType.ToLower())).Single() as ALMCustomizedType).Id;

            Common.SetDictionaryValue(ref requirementFields, "type-id", typeid);

            return base.Create(requirementFields);
        }
        #endregion
    }

    /// <summary>
    /// requirement class
    /// </summary>
    public class ALMRequirement : ALMEntity
    {
        #region private members
        #endregion       

        #region public Properties
        #endregion

        #region constructor
        #endregion

        #region public methods
        /// <summary>
        /// Add requirement under current requirement
        /// </summary>
        /// <param name="requirementType">Requirement type</param>
        /// <param name="requirementName">Requirement name</param>
        /// <param name="requirementFields">requirement fields which to be added.</param>
        /// <returns>Requirement</returns>
        public ALMRequirement AddRequirement(string requirementType, string requirementName, Dictionary<string, string> requirementFields)
        {
            if (string.IsNullOrEmpty(requirementType))
                throw new Exception("Requirement type cannot be empty.");

            if (string.IsNullOrEmpty(requirementName))
                throw new Exception("Requirement name cannot be empty.");

            ALMRequirements reqs = new ALMRequirements(Connection);
            if (reqs.CustomizedTypes.Where(o => o.Name.ToLower().Equals(requirementType.ToLower())).Count() == 0)
                throw new Exception(string.Format("Cannot find the requirement type: {0}, please double check", requirementType));
            string typeid = (reqs.CustomizedTypes.Where(o => o.Name.ToLower().Equals(requirementType.ToLower())).Single() as ALMCustomizedType).Id;

            Common.SetDictionaryValue(ref requirementFields, "type-id", typeid);
            Common.SetDictionaryValue(ref requirementFields, "parent-id", Id.ToString());
            Common.SetDictionaryValue(ref requirementFields, "owner", Connection.UserName);
            
            return new ALMRequirements(Connection).Create(requirementFields);
        }

        /// <summary>
        /// Update Requirement
        /// </summary>
        /// <returns></returns>
        public ALMRequirement Update()
        {
            if (Id <= 0) throw new Exception("Requirement Id cannot be empty.");

            return new ALMRequirements(Connection).Update(Id.Value, GetFieldsDictionary());
        }

        #region Links in defect
        /// <summary>
        /// Link defects to this requirements
        /// </summary>
        /// <param name="defectIds">defect ids</param>
        public void LinkToDefects(int[] defectIds)
        {
            if (defectIds == null)
                throw new Exception("Defect Ids are empty.");

            var defectlinks = new ALMDefectLinks(Connection);

            foreach (var defectId in defectIds)
            {
                defectlinks.Create(defectId, Id.Value, "requirement");
            }
        }

        /// <summary>
        /// Get all linked defects under the requirements
        /// </summary>
        /// <returns>list of defects</returns>
        public List<ALMDefect> GetLinkedDefects()
        {
            var linkeddefects = new List<ALMDefect>();

            var defectlinks = new ALMDefectLinks(Connection);
            var filters = new Dictionary<string, string>
                              {
                                  {"second-endpoint-id", Id.Value.ToString()},
                                  {"second-endpoint-type", "requirement"}
                              };
            var dllist = defectlinks.GetCollection(filters);

            if (dllist != null)
            {
                var defects = new ALMDefects(Connection);
                linkeddefects.AddRange(dllist.Select(dl => defects.Get(Convert.ToInt32(dl.GetField("first-endpoint-id").Value))).Where(d => d != null));
            }

            return linkeddefects;
        }

        /// <summary>
        /// delete defects under the requirement
        /// </summary>
        /// <param name="defectIds">defect ids</param>
        public void DeleteLinkedDefects(int[] defectIds)
        {
            if (defectIds == null)
                throw new Exception("Defect Ids are empty.");

            var defectlinks = new ALMDefectLinks(Connection);

            foreach (var dl in
                defectIds.Select(defectId => defectlinks.Get(defectId, Id.Value, "requirement")).Where(dl => dl != null))
            {
                dl.Delete();
            }
        }
        #endregion

        #endregion

        #region private methods
        #endregion
    }
}
